/*
 *   mod_auth_cert.c -- client certificate authentication module
 *
 *   Copyright (c) 2003-2006 by Michel Marti <mcmcfly@users.sf.net>
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * Trademarks are the property of their respective owners.
 *
 */

#include <httpd.h>
#include <http_core.h>
#include <http_config.h>
#include <http_protocol.h>
#include <ap_config.h>
#include <http_log.h>

#ifdef AP_SERVER_MAJORVERSION   /* Apache 2.0 */
module AP_MODULE_DECLARE_DATA cert_auth_module;
#include <ap_compat.h>
#if AP_SERVER_MINORVERSION_NUMBER < 2
#include <apr_compat.h>
#elif AP_SERVER_MINORVERSION_NUMBER == 2
#include <apr_strings.h>
#define ap_table_get apr_table_get
#define ap_table_set apr_table_set
#define ap_snprintf apr_snprintf
#define ap_pstrdup apr_pstrdup
#define ap_pcalloc apr_pcalloc
#endif
#include <apr_hooks.h>
#include <mod_ssl.h>
#define APLOG_STATUS  0,
#else                           /* Apache 1.3.x */
module MODULE_VAR_EXPORT cert_auth_module;
#define APLOG_STATUS
#define apr_pool_t pool
#define ap_configfile_t configfile_t
#define AP_INIT_TAKE1(directive, function, what, where, comment) \
  { directive, function, what, where, TAKE1, comment }
#define AP_INIT_FLAG(directive, function, what, where, comment) \
  { directive, function, what, where, FLAG, comment }
#define APR_OFFSETOF(a, b)	XtOffsetOf(a, b)
#endif

#define CERT_AUTH_TYPE "cert"

static char *mapUserDN(request_rec * r, char *userDN, char *mapfile);

typedef struct auth_config_struct {
	char *mapfile;
	int auth_authoritative;
	int set_auth_header;
} auth_config_rec;



static int cert2user(request_rec * r)
{
	char *user = NULL;
	char *userDN = NULL;

	auth_config_rec * sec = (auth_config_rec *)
		ap_get_module_config(r->per_dir_config, &cert_auth_module);

	/* Ignore request if we are not configured */
	if (sec == NULL || sec->mapfile == NULL) {
		return DECLINED;
	}

	/* ignore request if not authoritative, authorization header available and
	   auth-type != "cert" */
	if (!sec->auth_authoritative && ap_table_get(r->headers_in, "authorization") != NULL
			&& strcasecmp(ap_auth_type(r), CERT_AUTH_TYPE) != 0) {
		return DECLINED;
	}

	/* Get the clients subject DN from mod_ssl */

#ifdef AP_SERVER_MAJORVERSION
	static APR_OPTIONAL_FN_TYPE(ssl_var_lookup) *ssl_var_lookup;
	ssl_var_lookup = APR_RETRIEVE_OPTIONAL_FN(ssl_var_lookup);
	if (ssl_var_lookup) {
		userDN = ssl_var_lookup(r->pool, r->server, r->connection, r, "SSL_CLIENT_S_DN");
	}
#else
	ap_hook_call("ap::mod_ssl::var_lookup", &userDN, r->pool, r->server, r->connection, r, "SSL_CLIENT_S_DN");
#endif

	if (userDN == NULL || userDN[0] == 0) {
		ap_log_rerror(APLOG_MARK, (sec->auth_authoritative ? APLOG_ERR : APLOG_INFO)
				| APLOG_NOERRNO, APLOG_STATUS r, ": No client certificate.");

		if (!sec->auth_authoritative) {
			return DECLINED;
		}
		ap_note_basic_auth_failure(r);
		return HTTP_FORBIDDEN;
	}
	user = mapUserDN(r, userDN, sec->mapfile);
	if (user != NULL) {
		ap_log_rerror
			(APLOG_MARK, APLOG_INFO | APLOG_NOERRNO, APLOG_STATUS r, "User name mapped from Subject DN: '%s'.", user);

		/* Set the user and auth type */
#ifdef AP_SERVER_MAJORVERSION
		r->user = ap_pstrdup(r->pool, user);
		r->ap_auth_type = ap_pstrdup(r->pool, CERT_AUTH_TYPE);
#else
		r->connection->user = ap_pstrdup(r->pool, user);
		r->connection->ap_auth_type = ap_pstrdup(r->pool, CERT_AUTH_TYPE);
#endif

		return OK;
	}

	ap_log_rerror(APLOG_MARK, (sec->auth_authoritative ? APLOG_ERR : APLOG_INFO)
			| APLOG_NOERRNO, APLOG_STATUS r, ": No user mapping for Subject DN '%s'.", userDN);

	if (!sec->auth_authoritative) {
		return DECLINED;
	}
	ap_note_basic_auth_failure(r);
	return HTTP_FORBIDDEN;
}

static int set_auth_hdr(request_rec *r)
{
	auth_config_rec *sec = (auth_config_rec *)
		ap_get_module_config(r->per_dir_config, &cert_auth_module);
	/* Ignore request if we are not configured */
	if (sec == NULL || sec->mapfile == NULL) {
		return DECLINED;
	}

	/* set a fake authorization header containing username:password */
	if (sec->set_auth_header) {
		const char *authType = ap_auth_type(r);
		if (strcmp(authType, CERT_AUTH_TYPE) == 0) {
#ifdef AP_SERVER_MAJORVERSION
			char *user = r->user;
#else
			char *user = r->connection->user;
#endif
			if (user != NULL) {
				char auth[MAX_STRING_LEN];
				char *b64;

				ap_snprintf(auth, MAX_STRING_LEN, "%s:password", user);
				b64 = ap_pbase64encode(r->pool, auth);
				ap_snprintf(auth, MAX_STRING_LEN, "Basic %s", b64);
				ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, APLOG_STATUS r, "Setting 'authorization' header.");
				ap_table_set(r->headers_in, "Authorization", auth);
			}
		}
	}
	/* We don't really check access here */
	return DECLINED;
}

static char *mapUserDN(request_rec *r, char *userDN, char *mapfile)
{
	ap_configfile_t *f = NULL;
	char line[MAX_STRING_LEN];
	char *user_name;
	const char *ll;

#ifdef AP_SERVER_MAJORVERSION
	if (ap_pcfg_openfile(&f, r->pool, mapfile)!=APR_SUCCESS) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, APLOG_STATUS r, "Could not open map file: %s", mapfile);
		return NULL;
	}
#else
	f = ap_pcfg_openfile(r->pool, mapfile);
	if (f == NULL) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, APLOG_STATUS r, "Could not open map file: %s", mapfile);
		return NULL;
	}
#endif

	while (!(ap_cfg_getline(line, MAX_STRING_LEN, f))) {
		if ((line[0] == '#') || (!line[0])) {
			continue;
		}

		ll = line;
		user_name = ap_getword(r->pool, &ll, ':');

		if (strcmp(userDN, ll) == 0) {
			ap_cfg_closefile(f);
			return user_name;
		}
	}

	ap_cfg_closefile(f);
	return NULL;
}

static void *create_auth_dir_config(apr_pool_t *p, char *d)
{
    auth_config_rec *sec = (auth_config_rec *)
        ap_pcalloc(p, sizeof(auth_config_rec));
    sec->mapfile = NULL;
    sec->auth_authoritative = 1;
    sec->set_auth_header = 1;
    return sec;
}

static const command_rec auth_cmds[] = {
	AP_INIT_TAKE1("CertAuthMapFile",
			ap_set_file_slot,
			(void *) APR_OFFSETOF(auth_config_rec, mapfile),
			OR_AUTHCFG,
			"Text file containing user ID to Subject DN mappings"),
	AP_INIT_FLAG("CertAuthAuthoritative",
			ap_set_flag_slot,
			(void *) APR_OFFSETOF(auth_config_rec, auth_authoritative),
			OR_AUTHCFG,
			"Set to 'off' to allow access control to be passed along to "
			"lower modules if the UserID is not known to this module"),
	AP_INIT_FLAG("CertAuthSetAuthorization",
			ap_set_flag_slot,
			(void *) APR_OFFSETOF(auth_config_rec, set_auth_header),
			OR_AUTHCFG,
			"Set to 'off' if this module should not set a authorization header."),
	{NULL}
};


#ifndef STANDARD20_MODULE_STUFF

/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT cert_auth_module = {
	STANDARD_MODULE_STUFF,
	NULL,                       /* module initializer                  */
	create_auth_dir_config,     /* create per-dir    config structures */
	NULL,                       /* merge  per-dir    config structures */
	NULL,                       /* create per-server config structures */
	NULL,                       /* merge  per-server config structures */
	auth_cmds,                  /* table of config file commands       */
	NULL,                       /* [#8] MIME-typed-dispatched handlers */
	NULL,                       /* [#1] URI to filename translation    */
	cert2user,                  /* [#4] validate user id from request  */
	set_auth_hdr,               /* [#5] check if the user is ok _here_ */
	NULL,                       /* [#3] check access by host address   */
	NULL,                       /* [#6] determine MIME type            */
	NULL,                       /* [#7] pre-run fixups                 */
	NULL,                       /* [#9] log a transaction              */
	NULL,                       /* [#2] header parser                  */
	NULL,                       /* child_init                          */
	NULL,                       /* child_exit                          */
	NULL                        /* [#0] post read-request              */
#ifdef EAPI
	,NULL,                       /* EAPI: add_module                   */
	NULL,                       /* EAPI: remove_module                 */
	NULL,                       /* EAPI: rewrite_command               */
	NULL                        /* EAPI: new_connection                */
#endif
};

#else                           /* STANDARD20_MODULE_STUFF */

static void auth_register_hooks(apr_pool_t *p)
{
	static const char *aszPre[] = { "mod_ssl.c", NULL };
	ap_hook_check_user_id(cert2user, aszPre, NULL, APR_HOOK_FIRST);
	ap_hook_auth_checker(set_auth_hdr, aszPre, NULL, APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA cert_auth_module = {
	STANDARD20_MODULE_STUFF,
	create_auth_dir_config,     /* per-directory config creator */
	NULL,                       /* dir config merger */
	NULL,                       /* server config creator */
	NULL,                       /* server config merger */
	auth_cmds,                  /* command table */
	auth_register_hooks         /* register hooks */
};

#endif

